@@ -28,7 +28,8 @@
  * It handles the connection to a server and provides callbacks via pure virtual functions.
  *
  * To use the @b %NetMauMau @b Client @b API please read at first how to implement
- * NetMauMau::Client::AbstractClient. @n Useful functions you'll find in NetMauMau::Common.
+ * NetMauMau::Client::AbstractClient and @ref main.
+ * @n Useful functions you'll find in NetMauMau::Common.
  *
  * Link your client against @c -lnetmaumauclient and @c -lnetmaumaucommon
  *
@@ -44,20 +45,33 @@
 #include "clientconnection.h"
 #include "icard.h"
 
+#define MP_CNT_V05 23u
+#define MP_CNT_V07  3u
+#define MP_CNT_V08  1u
+#define MP_CNT_V13  1u
+
 namespace NetMauMau {
 
 /// @brief Classes and functions used by clients only
 namespace Client {
 
+template<class, std::size_t> class MappedMessageProcessor;
+
+struct _playInternalParams;
 class AbstractClientV05Impl;
 class IBase64;
 
 /**
+ * @ingroup main
+ *
  * @brief %Client interface to communicate with the server
  */
 class _EXPORT AbstractClientV05 : protected IPlayerPicListener {
 	DISALLOW_COPY_AND_ASSIGN(AbstractClientV05)
 
+	template<class, std::size_t> friend class MappedMessageProcessor;
+	template<class, class, std::size_t> friend struct MappedMessageAllocator;
+
 	friend class AbstractClientV05Impl;
 	friend class AbstractClientV07;
 	friend class AbstractClientV08;
@@ -65,6 +79,8 @@ class _EXPORT AbstractClientV05 : protected IPlayerPicListener {
 	friend class AbstractClientV11;
 	friend class AbstractClientV13;
 
+	typedef enum { OK, NOT_UNDERSTOOD, BREAK } PIRET;
+
 public:
 	/// @copydoc Connection::CAPABILITIES
 	typedef Connection::CAPABILITIES CAPABILITIES;
@@ -254,6 +270,9 @@ protected:
 	 */
 	explicit AbstractClientV05(const std::string &player, const std::string &server, uint16_t port);
 
+	explicit AbstractClientV05(const std::string &player, const std::string &server, uint16_t port,
+							   unsigned char sockopts);
+
 	/**
 	 * @brief Creates an @c AbstractClientV05 instance
 	 *
@@ -273,6 +292,10 @@ protected:
 	explicit AbstractClientV05(const std::string &player, const unsigned char *pngData,
 							   std::size_t pngDataLen, const std::string &server, uint16_t port);
 
+	explicit AbstractClientV05(const std::string &player, const unsigned char *pngData,
+							   std::size_t pngDataLen, const std::string &server, uint16_t port,
+							   unsigned char sockopts);
+
 	/**
 	 * @name Server requests
 	 * @{
@@ -547,26 +570,63 @@ protected:
 	virtual void unknownServerMessage(const std::string &msg) const = 0;
 
 private:
-	typedef enum { OK, NOT_UNDERSTOOD, BREAK } PIRET;
+	static bool isMisconfgMsg(const std::string &msg);
+	static bool isShutdownMsg(const std::string &msg);
+	static bool isLostConnMsg(const std::string &msg);
+	static std::string::size_type isRemotePlMsg(const std::string &msg);
 
-	virtual PIRET playInternal(std::string &msg, std::size_t *cturn, bool *initCardShown,
-							   std::string &cjackSuit,
-							   const NetMauMau::Common::ICard **lastPlayedCard)
+	void checkedError(const std::string &msg) const
 	throw(NetMauMau::Common::Exception::SocketException);
 
+	virtual PIRET playInternal(const _playInternalParams &p)
+	throw(NetMauMau::Common::Exception::SocketException);
+
+	PIRET performMessage(const _playInternalParams &) const;
+	PIRET performError(const _playInternalParams &) const
+	throw(NetMauMau::Common::Exception::SocketException);
+	PIRET performTurn(const _playInternalParams &) const;
+	PIRET performNextPlayer(const _playInternalParams &) const;
+	PIRET performStats(const _playInternalParams &) const;
+	PIRET performPlayerJoined(const _playInternalParams &) const;
+	PIRET performPlayerRejected(const _playInternalParams &) const;
+	PIRET performGetCards(const _playInternalParams &) const;
+	PIRET performInitialCard(const _playInternalParams &) const;
+	PIRET performTalonShuffled(const _playInternalParams &) const;
+	PIRET performOpenCard(const _playInternalParams &) const;
+	PIRET performPlayCard(const _playInternalParams &) const
+	throw(NetMauMau::Common::Exception::SocketException);
+	PIRET performSuspends(const _playInternalParams &) const;
+	PIRET performCardAccepted(const _playInternalParams &) const;
+	PIRET performCardRejected(const _playInternalParams &) const;
+	PIRET performCardCount(const _playInternalParams &) const;
+	PIRET performPlayedCard(const _playInternalParams &) const;
+	PIRET performJackSuit(const _playInternalParams &) const;
+	PIRET performJackModeOff(const _playInternalParams &) const;
+	PIRET performJackChoice(const _playInternalParams &) const;
+	PIRET performPlayerPicksCard(const _playInternalParams &) const;
+	PIRET performPlayerPicksCards(const _playInternalParams &) const;
+	PIRET performBye(const _playInternalParams &) const;
+
 private:
 	AbstractClientV05Impl *const _pimpl;
+	const MappedMessageProcessor<AbstractClientV05, MP_CNT_V05> *const m_mmp;
 };
 
 /**
+ * @ingroup main
+ *
  * @brief %Client interface to communicate with the server
  *
  * @since 0.7
  */
 class _EXPORT AbstractClientV07 : public AbstractClientV05 {
 	DISALLOW_COPY_AND_ASSIGN(AbstractClientV07)
+
+	template<class, class, std::size_t> friend struct MappedMessageAllocator;
+
 	friend class AbstractClientV08;
 	friend class AbstractClientV09;
+
 protected:
 	/**
 	 * @brief Creates an @c AbstractClientV07 instance
@@ -575,6 +635,9 @@ protected:
 	 */
 	explicit AbstractClientV07(const std::string &player, const std::string &server, uint16_t port);
 
+	explicit AbstractClientV07(const std::string &player, const std::string &server, uint16_t port,
+							   unsigned char sockopts);
+
 	/**
 	 * @brief Creates an @c AbstractClientV07 instance
 	 *
@@ -584,6 +647,11 @@ protected:
 	explicit AbstractClientV07(const std::string &player, const unsigned char *pngData,
 							   std::size_t pngDataLen, const std::string &server, uint16_t port);
 
+	explicit AbstractClientV07(const std::string &player, const unsigned char *pngData,
+							   std::size_t pngDataLen, const std::string &server, uint16_t port,
+							   unsigned char sockopts);
+
+
 	/**
 	 * @name Server requests
 	 *
@@ -637,20 +705,32 @@ public:
 private:
 	using AbstractClientV05::playInternal;
 
-	virtual AbstractClientV05::PIRET playInternal(std::string &msg, std::size_t *cturn,
-			bool *initCardShown, std::string &cjackSuit, const Common::ICard **lastPlayedCard)
+	virtual AbstractClientV05::PIRET playInternal(const _playInternalParams &p)
 	throw(NetMauMau::Common::Exception::SocketException);
+
+	PIRET performAceround(const _playInternalParams &) const;
+	PIRET performAceroundStarted(const _playInternalParams &) const;
+	PIRET performAceroundEnded(const _playInternalParams &) const;
+
+private:
+	const MappedMessageProcessor<AbstractClientV07, MP_CNT_V07> *const m_mmp;
 };
 
 /**
+ * @ingroup main
+ *
  * @brief %Client interface to communicate with the server
  *
  * @since 0.8
  */
 class _EXPORT AbstractClientV08 : public AbstractClientV07 {
 	DISALLOW_COPY_AND_ASSIGN(AbstractClientV08)
+
+	template<class, class, std::size_t> friend struct MappedMessageAllocator;
+
 	friend class AbstractClientV09;
 	friend class AbstractClientV13;
+
 protected:
 	using AbstractClientV07::playCard;
 
@@ -666,6 +746,9 @@ protected:
 	explicit AbstractClientV08(const std::string &player, const std::string &server, uint16_t port,
 							   uint32_t clientVersion);
 
+	explicit AbstractClientV08(const std::string &player, const std::string &server, uint16_t port,
+							   uint32_t clientVersion, unsigned char sockopts);
+
 	/**
 	 * @brief Creates an @c AbstractClientV08 instance
 	 *
@@ -680,6 +763,10 @@ protected:
 							   std::size_t pngDataLen, const std::string &server, uint16_t port,
 							   uint32_t clientVersion);
 
+	explicit AbstractClientV08(const std::string &player, const unsigned char *pngData,
+							   std::size_t pngDataLen, const std::string &server, uint16_t port,
+							   uint32_t clientVersion, unsigned char sockopts);
+
 	virtual Common::ICard *playCard(const CARDS &cards) const;
 
 	/**
@@ -704,13 +791,21 @@ protected:
 	virtual ~AbstractClientV08();
 
 private:
-	virtual PIRET playInternal(std::string &msg, std::size_t *cturn, bool *initCardShown,
-							   std::string &cjackSuit,
-							   const NetMauMau::Common::ICard **lastPlayedCard)
+	using AbstractClientV07::playInternal;
+
+	virtual PIRET playInternal(const _playInternalParams &p)
 	throw(NetMauMau::Common::Exception::SocketException);
+
+	PIRET performPlayCard(const _playInternalParams &) const
+	throw(NetMauMau::Common::Exception::SocketException);
+
+private:
+	const MappedMessageProcessor<AbstractClientV08, MP_CNT_V08> *const m_mmp;
 };
 
 /**
+ * @ingroup main
+ *
  * @brief %Client interface to communicate with the server
  *
  * @since 0.9
@@ -749,9 +844,6 @@ public:
 	 * @note if no scores are available (currently Windows servers) or the server never served
 	 * a game, the scores vector is empty
 	 *
-	 * @throws Exception::ScoresException if the scores couldn't get received or called within
-	 * a running game
-	 *
 	 * @param type type of scores
 	 * @param limit limit the result, @c 0 disables the limit
 	 * @param timeout throw a @c Exception::TimeoutException on exceeding @c timeout
@@ -782,6 +874,9 @@ protected:
 	explicit AbstractClientV09(const std::string &player, const std::string &server, uint16_t port,
 							   uint32_t clientVersion);
 
+	explicit AbstractClientV09(const std::string &player, const std::string &server, uint16_t port,
+							   uint32_t clientVersion, unsigned char sockopts);
+
 	/**
 	 * @brief Creates an @c AbstractClientV09 instance
 	 *
@@ -798,10 +893,16 @@ protected:
 							   std::size_t pngDataLen, const std::string &server, uint16_t port,
 							   uint32_t clientVersion);
 
+	explicit AbstractClientV09(const std::string &player, const unsigned char *pngData,
+							   std::size_t pngDataLen, const std::string &server, uint16_t port,
+							   uint32_t clientVersion, unsigned char sockopts);
+
 	virtual ~AbstractClientV09();
 };
 
 /**
+ * @ingroup main
+ *
  * @brief %Client interface to communicate with the server
  *
  * @since 0.11
@@ -809,6 +910,7 @@ protected:
 class _EXPORT AbstractClientV11 : public AbstractClientV09 {
 	DISALLOW_COPY_AND_ASSIGN(AbstractClientV11)
 public:
+	using AbstractClientV09::isPlayerImageUploadable;
 
 	/**
 	 * @name Server query methods
@@ -821,10 +923,13 @@ public:
 	 *
 	 * @param base64 interface to a custom implementation of @c IBase64
 	 *
+	 * @deprecated use
+	 * AbstractClientV05::isPlayerImageUploadable(const unsigned char *, std::size_t) instead
+	 *
 	 * @since 0.11
 	 */
 	static bool isPlayerImageUploadable(const unsigned char *pngData, std::size_t pngDataLen,
-										const IBase64 *base64);
+										const IBase64 *base64) _DEPRECATED;
 
 	/// @}
 
@@ -837,6 +942,9 @@ protected:
 	explicit AbstractClientV11(const std::string &player, const std::string &server, uint16_t port,
 							   uint32_t clientVersion);
 
+	explicit AbstractClientV11(const std::string &player, const std::string &server, uint16_t port,
+							   uint32_t clientVersion, unsigned char sockopts);
+
 	/**
 	 * @brief Creates an @c AbstractClientV11 instance
 	 *
@@ -844,10 +952,13 @@ protected:
 	 *
 	 * @param base64 interface to a custom implementation of @c IBase64
 	 *
+	 * @deprecated
+	 * use AbstractClientV09(const std::string &, const std::string &, uint16_t, uint32_t) instead
+	 *
 	 * @since 0.11
 	 */
 	explicit AbstractClientV11(const std::string &player, const std::string &server, uint16_t port,
-							   uint32_t clientVersion, const IBase64 *base64);
+							   uint32_t clientVersion, const IBase64 *base64) _DEPRECATED;
 
 	/**
 	 * @brief Creates an @c AbstractClientV11 instance
@@ -857,11 +968,14 @@ protected:
 	 *
 	 * @param base64 interface to a custom implementation of @c IBase64
 	 *
+	 * @deprecated use AbstractClientV09(const std::string &, const unsigned char *, std::size_t,
+	 * 									const std::string &, uint16_t, uint32_t) instead
+	 *
 	 * @since 0.11
 	 */
 	explicit AbstractClientV11(const std::string &player, const unsigned char *pngData,
 							   std::size_t pngDataLen, const std::string &server, uint16_t port,
-							   uint32_t clientVersion, const IBase64 *base64);
+							   uint32_t clientVersion, const IBase64 *base64) _DEPRECATED;
 
 	/**
 	 * @brief Creates an @c AbstractClientV11 instance
@@ -874,16 +988,23 @@ protected:
 					  std::size_t pngDataLen, const std::string &server, uint16_t port,
 					  uint32_t clientVersion);
 
+	AbstractClientV11(const std::string &player, const unsigned char *pngData,
+					  std::size_t pngDataLen, const std::string &server, uint16_t port,
+					  uint32_t clientVersion, unsigned char sockopts);
+
 	virtual ~AbstractClientV11();
 };
 
 /**
+ * @ingroup main
+ *
  * @brief %Client interface to communicate with the server
  *
  * @since 0.13
  */
 class _EXPORT AbstractClientV13 : public AbstractClientV11 {
 	DISALLOW_COPY_AND_ASSIGN(AbstractClientV13)
+	template<class, class, std::size_t> friend struct MappedMessageAllocator;
 protected:
 	/**
 	 * @brief Creates an @c AbstractClientV13 instance
@@ -893,6 +1014,9 @@ protected:
 	explicit AbstractClientV13(const std::string &player, const std::string &server, uint16_t port,
 							   uint32_t clientVersion);
 
+	explicit AbstractClientV13(const std::string &player, const std::string &server, uint16_t port,
+							   uint32_t clientVersion, unsigned char sockopts);
+
 	/**
 	 * @brief Creates an @c AbstractClientV13 instance
 	 *
@@ -900,10 +1024,13 @@ protected:
 	 *
 	 * @param base64 interface to a custom implementation of @c IBase64
 	 *
+	 * @deprecated
+	 * use AbstractClientV11(const std::string &, const std::string &, uint16_t, uint32_t) instead
+	 *
 	 * @since 0.13
 	 */
 	explicit AbstractClientV13(const std::string &player, const std::string &server, uint16_t port,
-							   uint32_t clientVersion, const IBase64 *base64);
+							   uint32_t clientVersion, const IBase64 *base64) _DEPRECATED;
 
 	/**
 	 * @brief Creates an @c AbstractClientV13 instance
@@ -913,11 +1040,14 @@ protected:
 	 *
 	 * @param base64 interface to a custom implementation of @c IBase64
 	 *
+	 * @deprecated use AbstractClientV11(const std::string &, const unsigned char *, std::size_t,
+	 * 									const std::string &, uint16_t, uint32_t) instead
+	 *
 	 * @since 0.13
 	 */
 	explicit AbstractClientV13(const std::string &player, const unsigned char *pngData,
 							   std::size_t pngDataLen, const std::string &server, uint16_t port,
-							   uint32_t clientVersion, const IBase64 *base64);
+							   uint32_t clientVersion, const IBase64 *base64) _DEPRECATED;
 
 	/**
 	 * @brief Creates an @c AbstractClientV13 instance
@@ -930,6 +1060,10 @@ protected:
 							   std::size_t pngDataLen, const std::string &server, uint16_t port,
 							   uint32_t clientVersion);
 
+	explicit AbstractClientV13(const std::string &player, const unsigned char *pngData,
+							   std::size_t pngDataLen, const std::string &server, uint16_t port,
+							   uint32_t clientVersion, unsigned char sockopts);
+
 	virtual ~AbstractClientV13();
 
 	/**
@@ -947,13 +1081,20 @@ protected:
 	/// @}
 
 private:
-	virtual PIRET playInternal(std::string &msg, std::size_t *cturn, bool *initCardShown,
-							   std::string &cjackSuit,
-							   const NetMauMau::Common::ICard **lastPlayedCard)
+	using AbstractClientV11::playInternal;
+
+	virtual PIRET playInternal(const _playInternalParams &p)
 	throw(NetMauMau::Common::Exception::SocketException);
+
+	PIRET performDirChange(const _playInternalParams &) const;
+
+private:
+	const MappedMessageProcessor<AbstractClientV13, MP_CNT_V13> *const m_mmp;
 };
 
 /**
+ * @ingroup main
+ *
  * @brief Alias to the current client interface to communicate with the server
  *
  * In your client subclass @c AbstractClient and implement all pure virtual methods. In the
@@ -966,10 +1107,6 @@ private:
  * A client can be obtained via git:
  * `git clone https://github.com/velnias75/NetMauMau-Qt-Client.git`
  *
- * @note All data is transferred as UTF-8 encoded byte strings
- * @note Starting with version 0.6 and with *libmagic* enabled, the server checks if the player
- * images are really in the PNG format
- *
  * @see AbstractClientV13
  *
  */
@@ -979,6 +1116,28 @@ typedef AbstractClientV13 AbstractClient;
 
 }
 
+/**
+ * @defgroup main Main Classes
+ * @brief The main classes of %NetMauMau
+ *
+ * The @c AbstractClientVx classes implement each the subset of the protocol up to the given
+ * version. They are each subclasses of the previous version. To get access to the @b latest
+ * version refer to the @em typedef NetMauMau::Client::AbstractClient. An important interface
+ * is NetMauMau::Common::ICard which describes a card and provides access to the @c ranks and
+ * @c suits.
+ *
+ * See @ref util for helper functions and macros.\n
+ * For error handling refer to @ref exceptions.
+ *
+ * @note All data is transferred as UTF-8 encoded byte strings
+ * @if libmagic @note This installation of the server enforces the player images to be in
+ * the PNG format @endif
+ * @if gsl @note This installation of the server uses the @b GNU @b Scientific @b Library
+ * for generating random numbers.\n
+ * See http://www.gnu.org/software/gsl/manual/html_node/Random-number-environment-variables.html
+ * or @em @c nmm-server(1) for more information.@endif
+ */
+
 #endif /* NETMAUMAU_ABSTRACTCLIENT_H */
 
 // kate: indent-mode cstyle; indent-width 4; replace-tabs off; tab-width 4; 
